import os
from pathlib import Path
from typing import List, Optional

from . import GlobalConfigRepository, BaseRepository
from ..models import ProjectData, Item, ItemWithProject, ItemState


class TaskRepository(BaseRepository):
    PROJECT_FILENAME = ".tasks.json"
    model = ProjectData

    def __init__(self, global_config_repository: GlobalConfigRepository, path: Optional[Path] = None) -> None:
        self._project_path = path or Path(os.getcwd())
        self._global_config_repository = global_config_repository
        self._global_config_changed = False
        super().__init__(self._project_path / self.PROJECT_FILENAME)

    def init(self):
        self._global_config_repository.add_project(self._project_path)
        self._global_config_repository.save()

    def list(self) -> List[Item]:
        return self._data.items

    def list_from_all_projects(self) -> List[ItemWithProject]:
        data = []

        for project in self._global_config_repository.get_all_projects():
            project_task_repository = TaskRepository(self._global_config_repository, Path(project))
            for item in project_task_repository.list():
                data.append(ItemWithProject(project=str(project), **item.dict()))

        return data

    def add(self, title: str, state: ItemState) -> None:
        new_item = Item(id=self._global_config_repository.get_items_counter(), state=state, title=title)
        self._data.items.append(new_item)
        self._global_config_changed = True

    def remove(self, task_id: int) -> None:
        self._data.items = [i for i in self._data.items if i.id != task_id]

    def save(self):
        super().save()

        if self._global_config_changed:
            self._global_config_repository.save()

    def _get_item_id_by_task_id(self, task_id: int) -> int:
        return next(i for i, item in enumerate(self._data.items) if item.id == task_id)

    def get_by_id(self, task_id: int) -> Item:
        return self._data.items[self._get_item_id_by_task_id(task_id)]

    def update(self, task_id: int, value: dict) -> None:
        item_id = self._get_item_id_by_task_id(task_id)
        origin_value = self._data.items[item_id]
        self._data.items[item_id] = Item(**{**origin_value.dict(), **value})



from typing import List, Optional

from ..models import Item
from ..time import get_time_difference


class ItemMixin:
    def _item_to_row(self, item: Item) -> List[str]:
        return [
            f"#{item.id}",
            item.title,
            self._get_tags(item.tags),
            str(item.state),
            get_time_difference(item.created_at),
        ]

    def _get_tags(self, tags: List[str]) -> str:
        return " ".join(map(lambda i: f"+{i}", tags))



from ..models import ItemWithProject
from .item_mixin import ItemMixin
from .list import ListView


class BasicListViewWithProject(ListView, ItemMixin):
    def _get_row(self, item: ItemWithProject) -> str:
        return " ".join(self._item_to_row(item) + [item.project])



from ..repositories import TaskRepository

class InitController:
    def __init__(self, task_repository: TaskRepository) -> None:
        self._task_repository = task_repository

    def init(self) -> None:
        try:
            self._task_repository.init()
            print("Init file already exists", ":sad_panda:")
        except:
            print("Tasks succesfully initialized", ":thumbs_up:")



from datetime import datetime


def _plurality(x: int) -> str:
    return "" if x == 1 else "s"


def get_time_difference(input_datetime: datetime) -> str:
    now = datetime.now()

    if (x := now.year - input_datetime.year) != 0:
        return f"{x} year{_plurality(x)} ago"

    if (x := now.month - input_datetime.month) != 0:
        return f"{x} month{_plurality(x)} ago"

    if (x := now.day - input_datetime.day) != 0:
        return f"{x} day{_plurality(x)} ago"

    if (x := now.hour - input_datetime.hour) != 0:
        return f"{x} hour{_plurality(x)} ago"

    if (x := now.minute - input_datetime.minute) != 0:
        return f"{x} minute{_plurality(x)} ago"

    if (x := now.second - input_datetime.second) != 0:
        return f"{x} second{_plurality(x)} ago"

    raise Exception("unexpected error")



import yaml

from ..editor import Editor
from ..repositories import TaskRepository, GlobalConfigRepository
from ..models import Configuration, ListViewType, ItemState
from ..view import BasicListView, BasicListViewWithProject, BasicTableView, BasicTableViewWithProjects, ListView

class TaskController:
    def __init__(self, task_repository: TaskRepository, global_config_repository: GlobalConfigRepository) -> None:
        self._task_repository = task_repository
        self._global_config_repository = global_config_repository

    def list(self, all_projects: bool) -> None:
        list_view = self._global_config_repository.get_list_view()

        if all_projects:
            data = self._task_repository.list_from_all_projects()

            if list_view == ListViewType.TABLE:
                view = BasicTableViewWithProjects(data)
            elif list_view == ListViewType.LIST:
                view = BasicListViewWithProject(data)
        else:
            data = self._task_repository.list()
            if list_view == ListViewType.TABLE:
                view = BasicTableView(data)
            elif list_view == ListViewType.LIST:
                view = BasicListView(data)

        view.print()

    def add(self, state: str, title: str) -> None:
        self._task_repository.add(" ".join(title), ItemState(state))
        self._task_repository.save()
        print("Task saved", ":vampire:")

    def edit(self, task_id: int) -> None:
        task = self._task_repository.get_by_id(task_id)
        editor = Editor(yaml.dump(task.dict(include={"title", "state", "tags"})))
        new_item_dict = yaml.safe_load(editor.edit())
        self._task_repository.update(task_id, new_item_dict)
        self._task_repository.save()

        print("Task edited")

    def remove(self, task_id: int) -> None:
        self._task_repository.remove(task_id)
        self._task_repository.save()
        print("Task removed :sad_panda:")



from __future__ import annotations

from datetime import datetime
from enum import Enum
from pathlib import Path
from typing import Dict, List, Optional

from pydantic import BaseModel, Field


class ItemIndex(BaseModel):
    id: int
    project: str


Tag = str


class ListViewType(str, Enum):
    TABLE = "table"
    LIST = "list"


class Configuration(BaseModel):
    list_view: ListViewType = ListViewType.TABLE

    class Config:
        use_enum_values = True


class GlobalConfig(BaseModel):
    all_projects: List[Path] = []
    items_counter: int = 0
    tags_index: Dict[Tag, ItemIndex] = {}
    config: Configuration = Configuration()

    def increase_items_counter(self) -> None:
        self.items_counter += 1


class ItemState(str, Enum):
    NEW = "NEW"
    TODO = "TODO"
    DONE = "DONE"


class Item(BaseModel):
    id: int
    created_at: datetime = Field(default_factory=datetime.now)
    updated_at: Optional[datetime] = None
    state: ItemState = ItemState.NEW
    tags: List[Tag] = []
    title: str

    class Config:
        use_enum_values = True


class ItemWithProject(Item):
    project: str


class ProjectData(BaseModel):
    created_at: datetime = Field(default_factory=datetime.now)
    items: List[Item] = []


DataPerProject = Dict[str, ProjectData]



from functools import wraps
from typing import Type

import click
import yaml
from rich import print
from inseminator import Container

from .controllers import InitController, TaskController
from .repositories import FileDoesntExistError, TaskRepository, GlobalConfigRepository


container = Container()

task_repository = container.resolve(TaskRepository)
global_config_repository = container.resolve(GlobalConfigRepository)
task_controller = container.resolve(TaskController)
init_controller = container.resolve(InitController)


class AliasedGroup(click.Group):
    def get_command(self, ctx, cmd_name):
        rv = click.Group.get_command(self, ctx, cmd_name)
        if rv is not None:
            return rv
        matches = [x for x in self.list_commands(ctx) if x.startswith(cmd_name)]
        if not matches:
            return None
        elif len(matches) == 1:
            return click.Group.get_command(self, ctx, matches[0])
        ctx.fail("Too many matches: %s" % ", ".join(sorted(matches)))


def catch_init_not_found(fn):
    @wraps(fn)
    def wrapper(*args, **kwargs):
        try:
            return fn(*args, **kwargs)
        except FileDoesntExistError:
            print("Project is not initialized :(")
            exit(1)

    return wrapper


@click.group(cls=AliasedGroup)
def cli() -> None:
    pass


@click.command()
def init() -> None:
    init_controller.init()


@click.command()
@click.option("-a", "--all-projects", default=False, type=bool, is_flag=True)
@catch_init_not_found
def list(all_projects: bool) -> None:
    task_controller.list(all_projects)


@click.command()
@click.confirmation_option(prompt="Are you sure you want to drop all your tasks there?")
@catch_init_not_found
def drop() -> None:
    task_repository.drop()
    print("Local tasks lost forever", ":sad_panda:")


@click.command()
@click.option("-s", "--state", default="NEW", type=str)
@click.argument("title", type=str, nargs=-1)
@catch_init_not_found
def add(state: str, title: str) -> None:
    task_controller.add(state, title)


@click.command()
@click.argument("task_id", type=int)
@catch_init_not_found
def remove(task_id: int) -> None:
    task_controller.remove(task_id)


@click.command()
@click.argument("task_id", type=int)
@catch_init_not_found
def edit(task_id: int) -> None:
    task_controller.edit(task_id)


@click.command()
@catch_init_not_found
def config() -> None:
    config = global_config_repository.get_config()
    print(yaml.dump(config.dict()))


@click.command()
@click.argument("name", type=str)
@click.argument("value", type=str)
@catch_init_not_found
def set_config(name: str, value: str) -> None:
    global_config_repository.set_config(name, value)


@click.command()
@catch_init_not_found
def projects() -> None:
    # TODO: move to controller
   for project in global_config_repository.get_all_projects():
        print(f" - {project}")


cli.add_command(add)
cli.add_command(edit)
cli.add_command(remove)
cli.add_command(list)
cli.add_command(init)
cli.add_command(drop)
cli.add_command(config)
cli.add_command(set_config)
cli.add_command(projects)



import json
from pathlib import Path
import os
from functools import cached_property
from typing import List

from . import BaseRepository
from ..models import GlobalConfig, Configuration, ListViewType


class GlobalConfigRepository(BaseRepository):
    GLOBAL_CONFIG_FILENANE = ".tasks.config.json"
    model = GlobalConfig

    def __init__(self) -> None:
        globa_config_filename = Path(os.path.expanduser("~")) / self.GLOBAL_CONFIG_FILENANE
        super().__init__(globa_config_filename)

    def add_project(self, project_path: Path) -> None:
        self._data.all_projects.append(str(project_path))

    def get_all_projects(self) -> List[Path]:
        return self._data.all_projects

    def register_project_if_possible(self, project_path: Path) -> None:
        if self._exists() and project_path not in self._data.all_projects:
            self._data.all_projects.append(str(project_path))
            self.save()

    def get_list_view(self) -> ListViewType:
        return self._data.config.list_view

    def set_config(self, name: str, value: str) -> None:
        config = self._data.config
        self._data.config = Configuration(**{**config.dict(), **{name: value}})
        self.save()

    def get_config(self) -> Configuration:
        return self._data.config

    def increase_items_counter(self):
        self._data.items_counter += 1

    def get_items_counter(self) -> int:
        return self._data.items_counter



import json
import os
from functools import cached_property
from pathlib import Path
from typing import List, Generic, Type, TypeVar
from abc import ABC, abstractmethod

from pydantic import BaseModel

from .errors import FileAlreadyExistsError, FileDoesntExistError
from ..models import ProjectData, GlobalConfig


T = TypeVar("T", ProjectData, GlobalConfig)


class BaseRepository(ABC, Generic[T]):
    @property
    @abstractmethod
    def model(self) -> Type[T]:
        ...

    def __init__(self, filename: Path) -> None:
        self._filename = filename

    @cached_property
    def _data(self) -> T:
        return self._load()

    def _exists(self) -> bool:
        return os.path.exists(self._filename)

    def _load(self) -> T:
        if not self._exists():
            raise FileDoesntExistError(self._filename)

        with open(self._filename, "r") as f:
            return self.model(**json.load(f))

    def _init(self, data: str) -> None:
        if self._exists():
            raise FileAlreadyExistsError(self._filename)

        with open(self._filename, "w+") as f:
            f.write(data)

    def save(self) -> None:
        with open(self._filename, "w") as f:
            f.write(self._data.json())

    def drop(self) -> None:
        if not self._exists():
            raise FileDoesntExistError(self._filename)

        os.remove(self._filename)



from typing import List

from ..models import Item
from .item_mixin import ItemMixin
from .list import ListView


class BasicListView(ListView, ItemMixin):
    def _get_row(self, item: Item) -> str:
        return " ".join(self._item_to_row(item))



from typing import List

from rich.table import Table

from ..models import ItemWithProject
from ..time import get_time_difference
from .item_mixin import ItemMixin
from .table import TableView


class BasicTableViewWithProjects(TableView, ItemMixin):
    def _create_table(self) -> Table:
        table = Table()
        table.add_column("ID", justify="right", style="cyan", no_wrap=True)
        table.add_column("Title", style="magenta")
        table.add_column("Tags", style="red")
        table.add_column("Status")
        table.add_column("Created", justify="right", style="green")
        table.add_column("Project", style="yellow")
        return table

    def _get_row(self, item: ItemWithProject) -> List[str]:
        row = super()._item_to_row(item)
        row.append(item.project)
        return row



from abc import ABC, abstractmethod
from typing import Generic, List, TypeVar

from rich import print
from rich.console import Console
from rich.table import Table

T = TypeVar("T")


class TableView(ABC, Generic[T]):
    def __init__(self, items: List[T]) -> None:
        self._items = items

    @abstractmethod
    def _create_table(self) -> Table:
        pass

    @abstractmethod
    def _get_row(self, item: T) -> List[str]:
        pass

    def print(self) -> None:
        console = Console()
        table = self._create_table()

        for item in self._items:
            table.add_row(*self._get_row(item))

        console.print(table)



from typing import List, Optional

from rich.table import Table

from ..models import Item
from .item_mixin import ItemMixin
from .table import TableView


class BasicTableView(TableView, ItemMixin):
    def _create_table(self) -> Table:
        table = Table()
        table.add_column("ID", justify="right", style="cyan", no_wrap=True)
        table.add_column("Title", style="magenta")
        table.add_column("Tags", style="red")
        table.add_column("Status")
        table.add_column("Created", justify="right", style="green")
        return table

    def _get_row(self, item: Item) -> List[str]:
        return super()._item_to_row(item)



from .basic_list_view import BasicListView
from .basic_list_view_with_projects import BasicListViewWithProject
from .basic_table_view import BasicTableView
from .basic_table_view_with_projects import BasicTableViewWithProjects
from .list import ListView
